from calendar import timegm
from datetime import datetime, timedelta, date, tzinfo as TZInfo
from typing import Callable, Self, Tuple, TypeVar, Generic, Mapping

T = TypeVar('T', bound=datetime | date)


class Queue(list[T]):

    delta: float
    delta_delta: float
    delta_type: str

    def __init__(self, delta: float | None, delta_delta: float, delta_type: str) -> None:
        super().__init__()
        if delta is None:
            self.delta = 0
            self.delta_delta = delta_delta
        else:
            self.delta = delta
            self.delta_delta = 0
        self.delta_type = delta_type

    def advance_next(self, delta: timedelta) -> None:
        self[-1] += delta  # type: ignore[assignment]

    def next(self) -> T:
        instance = self.pop(0)
        if not self:
            self.delta += self.delta_delta
            n = instance + timedelta(**{self.delta_type: self.delta})
            self.append(n)  # type: ignore[arg-type]
        return instance


class MockedCurrent(Generic[T]):

    _mock_queue: Queue[T]
    _mock_base_class: type
    _mock_class: type
    _mock_tzinfo: TZInfo | None
    _mock_date_type: type[date]
    _correct_mock_type: Callable[[T], Self] | None = None

    def __init_subclass__(
            cls,
            concrete: bool = False,
            queue: Queue[T] | None = None,
            strict: bool | None = None,
            tzinfo: TZInfo | None = None,
            date_type: type[date] = date,
    ) -> None:
        if concrete:
            assert not queue is None, 'queue must be passed if concrete=True'
            cls._mock_queue = queue
            cls._mock_base_class = cls.__bases__[0].__bases__[1]
            cls._mock_class = cls if strict else cls._mock_base_class
            cls._mock_tzinfo = tzinfo
            cls._mock_date_type = date_type

    @classmethod
    def add(cls, *args: int | T, **kw: int | TZInfo | None) -> None:
        if 'tzinfo' in kw or len(args) > 7:
            raise TypeError('Cannot add using tzinfo on %s' % cls.__name__)
        if args and isinstance(args[0], cls._mock_base_class):
            instance = args[0]
            instance_tzinfo = getattr(instance, 'tzinfo', None)
            if instance_tzinfo:
                if instance_tzinfo != cls._mock_tzinfo:
                    raise ValueError(
                        'Cannot add %s with tzinfo of %s as configured to use %s' % (
                            instance.__class__.__name__, instance_tzinfo, cls._mock_tzinfo
                        ))
                instance = instance.replace(tzinfo=None)  # type: ignore[attr-defined]
            if cls._correct_mock_type:
                instance = cls._correct_mock_type(instance)
        else:
            instance = cls(*args, **kw)  # type: ignore[arg-type]
        cls._mock_queue.append(instance)

    @classmethod
    def set(cls, *args: int | T, **kw: int | TZInfo | None) -> None:
        cls._mock_queue.clear()
        cls.add(*args, **kw)

    @classmethod
    def tick(cls, *args: timedelta, **kw: float) -> None:
        if kw:
            delta = timedelta(**kw)
        else:
            delta, = args
        cls._mock_queue.advance_next(delta)

    def __add__(self, other: timedelta) -> Self:
        instance = super().__add__(other)  # type: ignore[misc]
        if self._correct_mock_type:
            instance = self._correct_mock_type(instance)
        return instance

    def __new__(cls, *args: int, **kw: int | TZInfo | None) -> Self:
        if cls is cls._mock_class:
            return super().__new__(cls, *args, **kw)
        else:
            return cls._mock_class(*args, **kw)


def mock_factory(
        type_name: str,
        mock_class: type[MockedCurrent[T]],
        default: Tuple[int, ...],
        args: tuple[int | T | None | TZInfo, ...],
        kw: Mapping[str, int | TZInfo | None],
        delta: float | None,
        delta_type: str,
        delta_delta: float = 1,
        date_type: type[date] | None = None,
        tzinfo: TZInfo | None = None,
        strict: bool = False
) -> type[MockedCurrent[T]]:
    cls = type(
        type_name,
        (mock_class,),
        {},
        concrete=True,
        queue=Queue(delta, delta_delta, delta_type),
        strict=strict,
        tzinfo=tzinfo,
        date_type=date_type,
    )

    if args != (None,):
        if not (args or kw):
            args = default
        cls.add(*args, **kw)  # type: ignore[attr-defined]

    return cls


class MockDateTime(MockedCurrent[datetime], datetime):

    @classmethod
    def add(cls, *args: int | datetime, **kw: int | TZInfo | None) -> None:
        """
        This will add the :class:`datetime.datetime` created from the
        supplied parameters to the queue of datetimes to be returned by
        :meth:`~MockDateTime.now` or :meth:`~MockDateTime.utcnow`. An instance
        of :class:`~datetime.datetime` may also be passed as a single
        positional argument.
        """
        return super().add(*args, **kw)

    @classmethod
    def set(cls, *args: int | datetime, **kw: int | TZInfo | None) -> None:
        """
        This will set the :class:`datetime.datetime` created from the
        supplied parameters as the next datetime to be returned by
        :meth:`~MockDateTime.now` or :meth:`~MockDateTime.utcnow`, clearing out
        any datetimes in the queue.  An instance
        of :class:`~datetime.datetime` may also be passed as a single
        positional argument.
        """
        return super().set(*args, **kw)

    @classmethod
    def tick(cls, *args: timedelta, **kw: float) -> None:
        """
        This method should be called either with a :class:`~datetime.timedelta`
        as a positional argument, or with keyword parameters that will be used
        to construct a :class:`~datetime.timedelta`.

        The  :class:`~datetime.timedelta` will be used to advance the next datetime
        to be returned by :meth:`~MockDateTime.now` or :meth:`~MockDateTime.utcnow`.
        """
        return super().tick(*args, **kw)

    @classmethod  
    def _correct_mock_type(cls, instance: datetime) -> Self:
        return cls._mock_class(
            instance.year,
            instance.month,
            instance.day,
            instance.hour,
            instance.minute,
            instance.second,
            instance.microsecond,
            instance.tzinfo,
        )


    @classmethod
    def _adjust_instance_using_tzinfo(cls, instance: datetime) -> Self:
        if cls._mock_tzinfo:
            offset = cls._mock_tzinfo.utcoffset(instance)
            if offset is None:
                raise TypeError('tzinfo with .utcoffset() returning None is not supported')
            instance = instance - offset
        return instance  # type: ignore[return-value]

    @classmethod
    def now(cls, tz: TZInfo | None = None) -> Self:
        """
        :param tz: An optional timezone to apply to the returned time.
                If supplied, it must be an instance of a
                :class:`~datetime.tzinfo` subclass.

        This will return the next supplied or calculated datetime from the
        internal queue, rather than the actual current datetime.

        If `tz` is supplied, see :ref:`timezones`.
        """
        instance = cls._mock_queue.next()
        if tz is not None:
            instance = tz.fromutc(cls._adjust_instance_using_tzinfo(instance).replace(tzinfo=tz))
        return cls._correct_mock_type(instance)

    @classmethod
    def utcnow(cls) -> Self:
        """
        This will return the next supplied or calculated datetime from the
        internal queue, rather than the actual current UTC datetime.

        If you care about timezones, see :ref:`timezones`.
        """
        instance = cls._mock_queue.next()
        return cls._adjust_instance_using_tzinfo(instance)

    def date(self) -> date:
        """
        This will return the date component of the current mock instance,
        but using the date type supplied when the mock class was created.
        """
        return self._mock_date_type(
            self.year,
            self.month,
            self.day
            )


def mock_datetime(
        *args: int | datetime | None | TZInfo,
        tzinfo: TZInfo | None = None,
        delta: float | None = None,
        delta_type: str = 'seconds',
        date_type: type[date] = date,
        strict: bool = False,
        **kw: int | TZInfo | None,
) -> type[MockDateTime]:
    """
    .. currentmodule:: testfixtures.datetime

    A function that returns a mock object that can be used in place of
    the :class:`datetime.datetime` class but where the return value of
    :meth:`~MockDateTime.now` can be controlled.

    If a single positional argument of ``None`` is passed, then the
    queue of datetimes to be returned will be empty and you will need to
    call :meth:`~MockDateTime.set` or :meth:`~MockDateTime.add` before calling
    :meth:`~MockDateTime.now` or :meth:`~MockDateTime.utcnow`.

    If an instance of :class:`~datetime.datetime` is passed as a single
    positional argument, that will be used as the first date returned by
    :meth:`~MockDateTime.now`

    :param year:
      An optional year used to create the first datetime returned by :meth:`~MockDateTime.now`.

    :param month:
      An optional month used to create the first datetime returned by :meth:`~MockDateTime.now`.

    :param day:
      An optional day used to create the first datetime returned by :meth:`~MockDateTime.now`.

    :param hour:
      An optional hour used to create the first datetime returned by :meth:`~MockDateTime.now`.

    :param minute:
      An optional minute used to create the first datetime returned by :meth:`~MockDateTime.now`.

    :param second:
      An optional second used to create the first datetime returned by :meth:`~MockDateTime.now`.

    :param microsecond:
      An optional microsecond used to create the first datetime returned by
      :meth:`~MockDateTime.now`.

    :param tzinfo:
      An optional :class:`datetime.tzinfo`, see :ref:`timezones`.

    :param delta:
      The size of the delta to use between values returned from mocked class methods.
      If not specified, it will increase by 1 with each call to :meth:`~MockDateTime.now`.

    :param delta_type:
      The type of the delta to use between values returned from mocked class methods.
      This can be any keyword parameter accepted by the :class:`~datetime.timedelta` constructor.

    :param date_type:
      The type to use for the return value of the mocked class methods.
      This can help with gotchas that occur when type checking is performed on values returned
      by the :meth:`~testfixtures.datetime.MockDateTime.date` method.

    :param strict:
      If ``True``, calling the mock class and any of its methods will result in an instance of
      the mock being returned. If ``False``, the default, an instance of
      :class:`~datetime.datetime` will be returned instead.

    The mock returned will behave exactly as the :class:`datetime.datetime` class
    as well as being a subclass of :class:`~testfixtures.datetime.MockDateTime`.
    """
    if len(args) > 7:
        tzinfo = args[7]  # type: ignore[assignment]
        args = args[:7]
    else:
        tzinfo = tzinfo or (getattr(args[0], 'tzinfo', None) if args else None)
    return mock_factory(
        'MockDateTime',
        MockDateTime,
        (2001, 1, 1, 0, 0, 0),
        args,
        kw,
        tzinfo=tzinfo,
        delta=delta,
        delta_delta=10,
        delta_type=delta_type,
        date_type=date_type,
        strict=strict,
    )  # type: ignore[return-value]


class MockDate(MockedCurrent[date], date):

    @classmethod
    def _correct_mock_type(cls, instance: date) -> Self:
        return cls._mock_class(
            instance.year,
            instance.month,
            instance.day,
        )

    @classmethod
    def add(cls, *args: int | date, **kw: int | TZInfo | None) -> None:
        """
        This will add the :class:`datetime.date` created from the
        supplied parameters to the queue of dates to be returned by
        :meth:`~MockDate.today`.  An instance
        of :class:`~datetime.date` may also be passed as a single
        positional argument.
        """
        return super().add(*args, **kw)

    @classmethod
    def set(cls, *args: int | date, **kw: int | TZInfo | None) -> None:
        """
        This will set the :class:`datetime.date` created from the
        supplied parameters as the next date to be returned by
        :meth:`~MockDate.today`, regardless of any dates in the
        queue.   An instance
        of :class:`~datetime.date` may also be passed as a single
        positional argument.
        """
        return super().set(*args, **kw)

    @classmethod
    def tick(cls, *args: timedelta, **kw: float) -> None:
        """
        This method should be called either with a :class:`~datetime.timedelta`
        as a positional argument, or with keyword parameters that will be used
        to construct a :class:`~datetime.timedelta`.

        The  :class:`~datetime.timedelta` will be used to advance the next date
        to be returned by :meth:`~MockDate.today`.
        """
        return super().tick(*args, **kw)

    @classmethod
    def today(cls) -> Self:
        """
        This will return the next supplied or calculated date from the
        internal queue, rather than the actual current date.

        """
        return cls._mock_queue.next()  # type: ignore[return-value]


def mock_date(
        *args: int | date | None,
        delta: float | None = None,
        delta_type: str = 'days',
        strict: bool = False,
        **kw: int
) -> type[MockDate]:
    """
    .. currentmodule:: testfixtures.datetime

    A function that returns a mock object that can be used in place of
    the :class:`datetime.date` class but where the return value of
    :meth:`~datetime.date.today` can be controlled.

    If a single positional argument of ``None`` is passed, then the
    queue of dates to be returned will be empty and you will need to
    call :meth:`~MockDate.set` or :meth:`~MockDate.add` before calling
    :meth:`~MockDate.today`.

    If an instance of :class:`~datetime.date` is passed as a single
    positional argument, that will be used as the first date returned by
    :meth:`~datetime.date.today`

    :param year:
      An optional year used to create the first date returned by :meth:`~datetime.date.today`.

    :param month:
      An optional month used to create the first date returned by :meth:`~datetime.date.today`.

    :param day:
      An optional day used to create the first date returned by :meth:`~datetime.date.today`.

    :param delta:
      The size of the delta to use between values returned from :meth:`~datetime.date.today`.
      If not specified, it will increase by 1 with each call to :meth:`~datetime.date.today`.

    :param delta_type:
      The type of the delta to use between values returned from :meth:`~datetime.date.today`.
      This can be any keyword parameter accepted by the :class:`~datetime.timedelta` constructor.

    :param strict:
      If ``True``, calling the mock class and any of its methods will result in an instance of
      the mock being returned. If ``False``, the default, an instance of :class:`~datetime.date`
      will be returned instead.

    The mock returned will behave exactly as the :class:`datetime.date` class
    as well as being a subclass of :class:`~testfixtures.datetime.MockDate`.
    """
    return mock_factory(
        'MockDate', MockDate, (2001, 1, 1), args, kw,
        delta=delta,
        delta_type=delta_type,
        strict=strict,
    )  # type: ignore[return-value]


ms = 10**6


class MockTimeFactory(MockedCurrent[datetime], datetime):
    pass


class MockTime:

    def __init__(self, factory: type[MockedCurrent[datetime]]):
        self.factory = factory

    def add(self, *args: int | datetime, **kw: int | TZInfo | None) -> None:
        """
        This will add the time specified by the supplied parameters to the
        queue of times to be returned by calls to the mock. The
        parameters are the same as the :class:`datetime.datetime`
        constructor. An instance of :class:`~datetime.datetime` may also
        be passed as a single positional argument.
        """
        self.factory.add(*args, **kw)

    def set(self, *args: int | datetime, **kw: int | TZInfo | None) -> None:
        """
        This will set the time specified by the supplied parameters as
        the next time to be returned by a call to the mock, regardless of
        any times in the queue.  The parameters are the same as the
        :class:`datetime.datetime` constructor.  An instance of
        :class:`~datetime.datetime` may also be passed as a single
        positional argument.
        """
        self.factory.set(*args, **kw)

    def tick(self, *args: timedelta, **kw: float) -> None:
        """
        This method should be called either with a :class:`~datetime.timedelta`
        as a positional argument, or with keyword parameters that will be used
        to construct a :class:`~datetime.timedelta`.

        The  :class:`~datetime.timedelta` will be used to advance the next time
        to be returned by a call to the mock.
        """
        self.factory.tick(*args, **kw)

    def __call__(self) -> float:
        instance = self.factory._mock_queue.next()
        time: float = timegm(instance.utctimetuple())
        time += (float(instance.microsecond) / ms)
        return time


def mock_time(
        *args: int | datetime | None,
        delta: float | None = None,
        delta_type: str = 'seconds',
        **kw: int,
) -> MockTime:
    """
    .. currentmodule:: testfixtures.datetime

    A function that returns a :class:`mock object <testfixtures.datetime.MockTime>` that can be
    used in place of the :func:`time.time` function but where the return value can be
    controlled.

    If a single positional argument of ``None`` is passed, then the
    queue of times to be returned will be empty and you will need to
    call :meth:`~MockTime.set` or :meth:`~MockTime.add` before calling
    the mock.

    If an instance of :class:`~datetime.datetime` is passed as a single
    positional argument, that will be used to create the first time returned.

    :param year: An optional year used to create the first time returned.

    :param month: An optional month used to create the first time.

    :param day: An optional day used to create the first time.

    :param hour: An optional hour used to create the first time.

    :param minute: An optional minute used to create the first time.

    :param second: An optional second used to create the first time.

    :param microsecond: An optional microsecond used to create the first time.

    :param delta:
      The size of the delta to use between values returned.
      If not specified, it will increase by 1 with each call to the mock.

    :param delta_type:
      The type of the delta to use between values returned.
      This can be any keyword parameter accepted by the :class:`~datetime.timedelta` constructor.

    The :meth:`~testfixtures.datetime.MockTime.add`, :meth:`~testfixtures.datetime.MockTime.set`
    and :meth:`~testfixtures.datetime.MockTime.tick` methods on the mock can be used to
    control the return values.
    """
    if 'tzinfo' in kw or len(args) > 7 or (args and getattr(args[0], 'tzinfo', None)):
        raise TypeError("You don't want to use tzinfo with test_time")
    factory = mock_factory(
        'MockTime',
        MockTimeFactory,
        (2001, 1, 1, 0, 0, 0),
        args,
        kw,
        delta=delta,
        delta_type=delta_type,
    )
    return MockTime(factory)



import re

from testfixtures import StringComparison as S, compare
from unittest import TestCase


class Tests(TestCase):

    def test_equal_yes(self):
        self.assertTrue('on 40220' == S(r'on \d+'))

    def test_equal_no(self):
        self.assertFalse('on xxx' == S(r'on \d+'))

    def test_not_equal_yes(self):
        self.assertFalse('on 40220' != S(r'on \d+'))

    def test_not_equal_no(self):
        self.assertTrue('on xxx' != S(r'on \d+'))

    def test_comp_in_sequence(self):
        self.assertTrue((1, 2, 'on 40220') == (1, 2, S(r'on \d+')))

    def test_not_string(self):
        self.assertFalse(40220 == S(r'on \d+'))

    def test_repr(self):
        compare('<S:on \\d+>', repr(S(r'on \d+')))

    def test_str(self):
        compare('<S:on \\d+>', str(S(r'on \d+')))

    def test_sort(self):
        a = S('a')
        b = S('b')
        c = S('c')
        compare(sorted(('d', c, 'e', a, 'a1', b)), expected=[a, 'a1', b, c, 'd', 'e'])

    def test_flags_argument(self):
        compare(S(".*bar", re.DOTALL), actual="foo\nbar")

    def test_flags_parameter(self):
        compare(S(".*bar", flags=re.DOTALL), actual="foo\nbar")

    def test_flags_names(self):
        compare(S(".*BaR", dotall=True, ignorecase=True), actual="foo\nbar")



import sys
from unittest import TestCase

from testfixtures import Comparison as C, TempDirectory, diff, Comparison
from testfixtures.tests.sample1 import SampleClassA, a_function


class AClass:

    def __init__(self, x, y=None):
        self.x = x
        if y:
            self.y = y

    def __repr__(self):
        return '<'+self.__class__.__name__+'>'


class BClass(AClass):
    pass


class WeirdException(Exception):
    def __init__(self, x, y):
        self.x = x
        self.y = y


class X:
    __slots__ = ['x']


class FussyDefineComparison:

    def __init__(self, attr):
        self.attr = attr

    def __eq__(self, other):
        if not isinstance(other, self.__class__):  # pragma: no cover
            raise TypeError()
        return False  # pragma: no cover

    def __ne__(self, other):
        return not self == other  # pragma: no cover


def compare_repr(obj, expected):
    actual = diff(expected, repr(obj))
    if actual:  # pragma: no cover
        raise AssertionError(actual)


class TestC(TestCase):

    def test_example(self):
        # In this pattern, we want to check a sequence is
        # of the correct type and order.
        r = a_function()
        self.assertEqual(r, (
            C('testfixtures.tests.sample1.SampleClassA'),
            C('testfixtures.tests.sample1.SampleClassB'),
            C('testfixtures.tests.sample1.SampleClassA'),
            ))
        # We also want to check specific parts of some
        # of the returned objects' attributes
        self.assertEqual(r[0].args[0], 1)
        self.assertEqual(r[1].args[0], 2)
        self.assertEqual(r[2].args[0], 3)

    def test_example_with_object(self):
        # Here we see compare an object with a Comparison
        # based on an object of the same type and with the
        # same attributes:
        self.assertEqual(
            C(AClass(1, 2)),
            AClass(1, 2),
            )
        # ...even though the original class doesn't support
        # meaningful comparison:
        self.assertNotEqual(
            AClass(1, 2),
            AClass(1, 2),
            )

    def test_example_with_vars(self):
        # Here we use a Comparison to make sure both the
        # type and attributes of an object are correct.
        self.assertEqual(
            C('testfixtures.tests.test_comparison.AClass',
              x=1, y=2),
            AClass(1, 2),
            )

    def test_example_with_odd_vars(self):
        # If the variable names class with parameters to the
        # Comparison constructor, they can be specified in a
        # dict:
        self.assertEqual(
            C('testfixtures.tests.test_comparison.AClass',
              {'x': 1, 'y': 2}),
            AClass(1, 2),
            )

    def test_example_partial(self):
        self.assertEqual(
            C('testfixtures.tests.test_comparison.AClass',
              x=1,
              partial=True),
            AClass(1, 2),
            )

    def test_example_dont_use_c_wrappers_on_both_sides(self):
        # NB: don't use C wrappers on both sides!
        e = ValueError('some message')
        x, y = C(e), C(e)
        assert x != y
        compare_repr(x, "<C:builtins.ValueError(failed)>wrong type</>")
        compare_repr(y, "<C:builtins.ValueError>args: ('some message',)</>")

    def test_repr_module(self):
        compare_repr(C('datetime'), '<C:datetime>')

    def test_repr_class(self):
        compare_repr(C('testfixtures.tests.sample1.SampleClassA'),
                     '<C:testfixtures.tests.sample1.SampleClassA>')

    def test_repr_function(self):
        compare_repr(C('testfixtures.tests.sample1.z'),
                     '<C:testfixtures.tests.sample1.z>')

    def test_repr_instance(self):
        compare_repr(C(SampleClassA('something')),
                     "<C:testfixtures.tests.sample1.SampleClassA>"
                     "args: ('something',)"
                     "</>"
                     )

    def test_repr_exception(self):
        compare_repr(C(ValueError('something')), "<C:builtins.ValueError>args: ('something',)</>")

    def test_repr_exception_not_args(self):
        compare_repr(
            C(WeirdException(1, 2)),
            "\n<C:testfixtures.tests.test_comparison.WeirdException>\n"
            "args: (1, 2)\n"
            "x: 1\n"
            "y: 2\n"
            "</C:testfixtures.tests.test_comparison.WeirdException>"
        )

    def test_repr_class_and_vars(self):
        compare_repr(
            C(SampleClassA, {'args': (1,)}),
            "<C:testfixtures.tests.sample1.SampleClassA>args: (1,)</>"
        )

    def test_repr_nested(self):
        compare_repr(
            C(SampleClassA, y=C(AClass), z=C(BClass(1, 2))),
            "\n"
            "<C:testfixtures.tests.sample1.SampleClassA>\n"
            "y: <C:testfixtures.tests.test_comparison.AClass>\n"
            "z: \n"
            "  <C:testfixtures.tests.test_comparison.BClass>\n"
            "  x: 1\n"
            "  y: 2\n"
            "  </C:testfixtures.tests.test_comparison.BClass>\n"
            "</C:testfixtures.tests.sample1.SampleClassA>"
            )

    def test_repr_failed_wrong_class(self):
        c = C('testfixtures.tests.test_comparison.AClass', x=1, y=2)
        assert c != BClass(1, 2)
        compare_repr(c,
                     "<C:testfixtures.tests.test_comparison.AClass(failed)>"
                     "wrong type</>"
                     )

    def test_repr_failed_all_reasons_in_one(self):
        c = C('testfixtures.tests.test_comparison.AClass',
              y=5, z='missing')
        assert c != AClass(1, 2)
        compare_repr(c,
                     "\n"
                     "<C:testfixtures.tests.test_comparison.AClass(failed)>\n"
                     "attributes in Comparison but not actual:\n"
                     "'z': 'missing'\n\n"
                     "attributes in actual but not Comparison:\n"
                     "'x': 1\n\n"
                     "attributes differ:\n"
                     "'y': 5 (Comparison) != 2 (actual)\n"
                     "</C:testfixtures.tests.test_comparison.AClass>",
                     )

    def test_repr_failed_not_in_other(self):
        c = C('testfixtures.tests.test_comparison.AClass',
              x=1, y=2, z=(3, ))
        assert c != AClass(1, 2)
        compare_repr(c ,
                     "\n"
                     "<C:testfixtures.tests.test_comparison.AClass(failed)>\n"
                     "attributes same:\n"
                     "['x', 'y']\n\n"
                     "attributes in Comparison but not actual:\n"
                     "'z': (3,)\n"
                     "</C:testfixtures.tests.test_comparison.AClass>",
                     )

    def test_repr_failed_not_in_self(self):
        c = C('testfixtures.tests.test_comparison.AClass', y=2)
        assert c != AClass(x=(1, ), y=2)
        compare_repr(c,
                     "\n"
                     "<C:testfixtures.tests.test_comparison.AClass(failed)>\n"
                     "attributes same:\n"
                     "['y']\n\n"
                     "attributes in actual but not Comparison:\n"
                     "'x': (1,)\n"
                     "</C:testfixtures.tests.test_comparison.AClass>",
                     )

    def test_repr_failed_not_in_self_partial(self):
        c = C('testfixtures.tests.test_comparison.AClass', x=1, y=2, z=(3, ), partial=True)
        assert c != AClass(x=1, y=2)
        compare_repr(c,
                     "\n"
                     "<C:testfixtures.tests.test_comparison.AClass(failed)>\n"
                     "attributes same:\n"
                     "['x', 'y']\n\n"
                     "attributes in Comparison but not actual:\n"
                     "'z': (3,)\n"
                     "</C:testfixtures.tests.test_comparison.AClass>",
                     )

    def test_repr_failed_one_attribute_not_equal(self):
        c = C('testfixtures.tests.test_comparison.AClass', x=1, y=(2, ))
        assert c != AClass(1, (3, ))
        compare_repr(c,
                     "\n"
                     "<C:testfixtures.tests.test_comparison.AClass(failed)>\n"
                     "attributes same:\n"
                     "['x']\n\n"
                     "attributes differ:\n"
                     "'y': (2,) (Comparison) != (3,) (actual)\n"
                     "</C:testfixtures.tests.test_comparison.AClass>",
                     )

    def test_repr_failed_nested(self):
        left_side = [C(AClass, x=1, y=2),
                     C(BClass, x=C(AClass, x=1, y=2), y=C(AClass))]
        right_side = [AClass(1, 3), AClass(1, 3)]

        # do the comparison
        left_side == right_side

        compare_repr(
            left_side,
            "[\n"
            "<C:testfixtures.tests.test_comparison.AClass(failed)>\n"
            "attributes same:\n"
            "['x']\n\n"
            "attributes differ:\n"
            "'y': 2 (Comparison) != 3 (actual)\n"
            "</C:testfixtures.tests.test_comparison.AClass>, \n"
            "<C:testfixtures.tests.test_comparison.BClass>\n"
            "x: \n"
            "  <C:testfixtures.tests.test_comparison.AClass>\n"
            "  x: 1\n"
            "  y: 2\n"
            "  </C:testfixtures.tests.test_comparison.AClass>\n"
            "y: <C:testfixtures.tests.test_comparison.AClass>\n"
            "</C:testfixtures.tests.test_comparison.BClass>]"
        )

        compare_repr(right_side, "[<AClass>, <AClass>]")

    def test_repr_failed_nested_failed(self):
        left_side = [C(AClass, x=1, y=2),
                     C(BClass,
                       x=C(AClass, x=1, partial=True),
                       y=C(AClass, z=2))]
        right_side = [AClass(1, 2),
                      BClass(AClass(1, 2), AClass(1, 2))]

        # do the comparison
        left_side == right_side

        compare_repr(
            left_side,
            "[\n"
            "<C:testfixtures.tests.test_comparison.AClass>\n"
            "x: 1\n"
            "y: 2\n"
            "</C:testfixtures.tests.test_comparison.AClass>, \n"
            "<C:testfixtures.tests.test_comparison.BClass(failed)>\n"
            "attributes same:\n"
            "['x']\n\n"
            "attributes differ:\n"
            "'y': \n"
            "<C:testfixtures.tests.test_comparison.AClass(failed)>\n"
            "attributes in Comparison but not actual:\n"
            "'z': 2\n\n"
            "attributes in actual but not Comparison:\n"
            "'x': 1\n"
            "'y': 2\n"
            "</C:testfixtures.tests.test_comparison.AClass> (Comparison) != <AClass> (actual)\n"
            "</C:testfixtures.tests.test_comparison.BClass>]",
        )

        compare_repr(right_side, '[<AClass>, <BClass>]')

    def test_repr_failed_passed_failed(self):
        c = C('testfixtures.tests.test_comparison.AClass', x=1, y=2)
        assert c != AClass(1, 3)
        compare_repr(c,
                     "\n"
                     "<C:testfixtures.tests.test_comparison.AClass(failed)>\n"
                     "attributes same:\n"
                     "['x']\n\n"
                     "attributes differ:\n"
                     "'y': 2 (Comparison) != 3 (actual)\n"
                     "</C:testfixtures.tests.test_comparison.AClass>",
                     )

        assert c == AClass(1, 2)

        assert c != AClass(3, 2)
        compare_repr(c,
                     "\n"
                     "<C:testfixtures.tests.test_comparison.AClass(failed)>\n"
                     "attributes same:\n"
                     "['y']\n\n"
                     "attributes differ:\n"
                     "'x': 1 (Comparison) != 3 (actual)\n"
                     "</C:testfixtures.tests.test_comparison.AClass>",
                     )

    def test_first(self):
        self.assertEqual(
            C('testfixtures.tests.sample1.SampleClassA'),
            SampleClassA()
            )

    def test_second(self):
        self.assertEqual(
            SampleClassA(),
            C('testfixtures.tests.sample1.SampleClassA'),
            )

    def test_not_same_first(self):
        self.assertNotEqual(
            C('datetime'),
            SampleClassA()
            )

    def test_not_same_second(self):
        self.assertNotEqual(
            SampleClassA(),
            C('datetime')
            )

    def test_object_supplied(self):
        self.assertEqual(
            SampleClassA(1),
            C(SampleClassA(1))
            )

    def test_class_and_vars(self):
        self.assertEqual(
            SampleClassA(1),
            C(SampleClassA, {'args': (1,)})
            )

    def test_class_and_kw(self):
        self.assertEqual(
            SampleClassA(1),
            C(SampleClassA, args=(1,))
            )

    def test_class_and_vars_and_kw(self):
        self.assertEqual(
            AClass(1, 2),
            C(AClass, {'x': 1}, y=2)
            )

    def test_object_and_vars(self):
        # vars passed are used instead of the object's
        self.assertEqual(
            SampleClassA(1),
            C(SampleClassA(), {'args': (1,)})
            )

    def test_object_and_kw(self):
        # kws passed are used instead of the object's
        self.assertEqual(
            SampleClassA(1),
            C(SampleClassA(), args=(1,))
            )

    def test_object_partial(self):
        # only attributes on comparison object
        # are used
        self.assertEqual(
            C(AClass(1), partial=True),
            AClass(1, 2),
            )

    def run_property_equal_test(self, partial):
        class SomeClass:
            @property
            def prop(self):
                return 1

        self.assertEqual(
            C(SomeClass, prop=1, partial=partial),
            SomeClass()
        )

    def test_property_equal(self):
        self.run_property_equal_test(partial=False)

    def test_property_equal_partial(self):
        self.run_property_equal_test(partial=True)

    def run_property_not_equal_test(self, partial):
        class SomeClass:
            @property
            def prop(self):
                return 1

        c = C(SomeClass, prop=2, partial=partial)
        self.assertNotEqual(c, SomeClass())
        compare_repr(
            c,
            "\n"
            "<C:testfixtures.tests.test_comparison.SomeClass(failed)>\n"
            "attributes differ:\n"
            "'prop': 2 (Comparison) != 1 (actual)\n"
            "</C:testfixtures.tests.test_comparison.SomeClass>")

    def test_property_not_equal(self):
        self.run_property_not_equal_test(partial=False)

    def test_property_not_equal_partial(self):
        self.run_property_not_equal_test(partial=True)

    def run_method_equal_test(self, partial):
        class SomeClass:
            def method(self):
                pass  # pragma: no cover

        instance = SomeClass()
        self.assertEqual(
            C(SomeClass, method=instance.method, partial=partial),
            instance
        )

    def test_method_equal(self):
        self.run_method_equal_test(partial=False)

    def test_method_equal_partial(self):
        self.run_method_equal_test(partial=True)

    def run_method_not_equal_test(self, partial):
        class SomeClass: pass
        instance = SomeClass()
        instance.method = min

        c = C(SomeClass, method=max, partial=partial)
        self.assertNotEqual(c, instance)
        compare_repr(
            c,
            "\n"
            "<C:testfixtures.tests.test_comparison.SomeClass(failed)>\n"
            "attributes differ:\n"
            "'method': <built-in function max> (Comparison)"
            " != <built-in function min> (actual)\n"
            "</C:testfixtures.tests.test_comparison.SomeClass>"
        )

    def test_method_not_equal(self):
        self.run_method_not_equal_test(partial=False)

    def test_method_not_equal_partial(self):
        self.run_method_not_equal_test(partial=True)

    def test_exception(self):
        self.assertEqual(
            ValueError('foo'),
            C(ValueError('foo'))
            )

    def test_exception_class_and_args(self):
        self.assertEqual(
            ValueError('foo'),
            C(ValueError, args=('foo', ))
            )

    def test_exception_instance_and_args(self):
        self.assertEqual(
            ValueError('foo'),
            C(ValueError('bar'), args=('foo', ))
            )

    def test_exception_not_same(self):
        self.assertNotEqual(
            ValueError('foo'),
            C(ValueError('bar'))
            )

    def test_exception_no_args_different(self):
        self.assertNotEqual(
            WeirdException(1, 2),
            C(WeirdException(1, 3))
            )

    def test_exception_no_args_same(self):
        self.assertEqual(
            C(WeirdException(1, 2)),
            WeirdException(1, 2)
            )

    def test_repr_file_different(self):
        with TempDirectory() as d:
            path = d.write('file', b'stuff')
            f = open(path)
            f.close()
        c = C('io.TextIOWrapper', name=path, mode='r', closed=False,
              partial=True)
        assert f != c
        compare_repr(c,
                     "\n"
                     "<C:_io.TextIOWrapper(failed)>\n"
                     "attributes same:\n"
                     "['mode', 'name']\n\n"
                     "attributes differ:\n"
                     "'closed': False (Comparison) != True (actual)\n"
                     "</C:_io.TextIOWrapper>",
                     )

    def test_file_same(self):
        with TempDirectory() as d:
            path = d.write('file', b'stuff')
            f = open(path)
            f.close()
        self.assertEqual(
            f,
            C('io.TextIOWrapper', name=path, mode='r', closed=True,
              partial=True)
            )

    def test_no___dict___strict(self):
        c = C(X, x=1)
        assert c != X()
        compare_repr(c, "\n"
                        "<C:testfixtures.tests.test_comparison.X(failed)>\n"
                        "attributes in Comparison but not actual:\n"
                        "'x': 1\n"
                        "</C:testfixtures.tests.test_comparison.X>")

    def test_no___dict___partial_same(self):
        x = X()
        x.x = 1
        self.assertEqual(C(X, x=1, partial=True), x)

    def test_no___dict___partial_missing_attr(self):
        c = C(X, x=1, partial=True)
        assert c != X()
        compare_repr(c,
                     "\n"
                     "<C:testfixtures.tests.test_comparison.X(failed)>\n"
                     "attributes in Comparison but not actual:\n"
                     "'x': 1\n"
                     "</C:testfixtures.tests.test_comparison.X>",
                     )

    def test_no___dict___partial_different(self):
        x = X()
        x.x = 2
        c = C(X, x=1, y=2, partial=True)
        assert c != x
        compare_repr(c,
                     "\n"
                     "<C:testfixtures.tests.test_comparison.X(failed)>\n"
                     "attributes in Comparison but not actual:\n"
                     "'y': 2\n\n"
                     "attributes differ:\n"
                     "'x': 1 (Comparison) != 2 (actual)\n"
                     "</C:testfixtures.tests.test_comparison.X>",
                     )

    def test_compared_object_defines_eq(self):
        # If an object defines eq, such as Django instances,
        # things become tricky

        class Annoying:
            def __init__(self):
                self.eq_called = 0

            def __eq__(self, other):
                self.eq_called += 1
                if isinstance(other, Annoying):
                    return True
                return False

        self.assertEqual(Annoying(), Annoying())

        # Suddenly, order matters.

        # This order is wrong, as it uses the class's __eq__:
        self.assertFalse(Annoying() == C(Annoying))
        # but on __eq__ is used as a fallback:
        self.assertTrue(Annoying() != C(Annoying))

        # This is the right ordering:
        self.assertTrue(C(Annoying) == Annoying())
        self.assertFalse(C(Annoying) != Annoying())

        # When the ordering is right, you still get the useful
        # comparison representation afterwards
        c = C(Annoying, eq_called=1)
        c == Annoying()
        compare_repr(
            c,
            '\n<C:testfixtures.tests.test_comparison.Annoying(failed)>\n'
            'attributes differ:\n'
            "'eq_called': 1 (Comparison) != 0 (actual)\n"
            '</C:testfixtures.tests.test_comparison.Annoying>'
        )

    def test_importerror(self):
        assert C(ImportError('x')) == ImportError('x')

    def test_class_defines_comparison_strictly(self):
        self.assertEqual(
            C('testfixtures.tests.test_comparison.FussyDefineComparison',
              attr=1),
            FussyDefineComparison(1)
            )

    def test_cant_resolve(self):
        try:
            C('testfixtures.bonkers')
        except Exception as e:
            self.assertTrue(isinstance(e, AttributeError))
            self.assertEqual(
                e.args,
                ("'testfixtures.bonkers' could not be resolved", )
                )
        else:
            self.fail('No exception raised!')

    def test_no_name(self):
        class NoName:
            pass
        NoName.__name__ = ''
        NoName.__module__ = ''
        c = C(NoName)
        self.assertEqual(repr(c), "<C:<class '.TestC.test_no_name.<locals>.NoName'>>")

    def test_missing_expected_attribute_not_partial(self):

        class MyClass:
            def __init__(self, **attrs):
                self.__dict__.update(attrs)

        c = Comparison(MyClass, b=2, c=3, partial=False)
        assert c != MyClass(a=1, b=2)

    def test_missing_expected_attribute_partial(self):

        class MyClass:
            def __init__(self, **attrs):
                self.__dict__.update(attrs)

        c = Comparison(MyClass, b=2, c=3, partial=True)
        assert c != MyClass(a=1, b=2)

    def test_extra_expected_attribute_not_partial(self):

        class MyClass:
            def __init__(self, **attrs):
                self.__dict__.update(attrs)

        c = Comparison(MyClass, a=1, partial=False)
        assert c != MyClass(a=1, b=2)

    def test_extra_expected_attribute_partial(self):

        class MyClass:
            def __init__(self, **attrs):
                self.__dict__.update(attrs)

        c = Comparison(MyClass, a=1, partial=True)
        assert c == MyClass(a=1, b=2)



from typing import Any, Sequence, Iterable

from django.db.models import Model, Field

from . import compare as base_compare
from .comparison import _compare_mapping, register, CompareContext, unspecified, Registry


def instance_fields(instance: Model) -> Iterable[Field]:
    opts = instance._meta
    for name in (
        'concrete_fields',
        'virtual_fields',
        'private_fields',
    ):
        fields = getattr(opts, name, None)
        if fields:
            for field in fields:
                yield field


def model_to_dict(
        instance: Model,
        exclude: Sequence[str],
        include_not_editable: bool,
) -> dict[str, Any]:
    data = {}
    for f in instance_fields(instance):
        if f.name in exclude:
            continue
        if not getattr(f, 'editable', False) and not include_not_editable:
            continue
        data[f.name] = f.value_from_object(instance)
    return data


def compare_model(x: Model, y: Model, context: CompareContext) -> str | None:
    """
    Returns an informative string describing the differences between the two
    supplied Django model instances. The way in which this comparison is
    performed can be controlled using the following parameters:

    :param ignore_fields:
      A sequence of fields to ignore during comparison, most commonly
      set to ``['id']``. By default, no fields are ignored.

    :param non_editable_fields:
      If `True`, then fields with ``editable=False`` will be included in the
      comparison. By default, these fields are ignored.
    """
    ignore_fields = context.get_option('ignore_fields', set())
    non_editable_fields= context.get_option('non_editable_fields', False)
    args: Any = []
    for obj in x, y:
        args.append(model_to_dict(obj, ignore_fields, non_editable_fields))
    args.append(context)
    args.append(x)
    return _compare_mapping(*args)


register(Model, compare_model)


def compare(
        *args: Any,
        x: Any = unspecified,
        y: Any = unspecified,
        expected: Any = unspecified,
        actual: Any = unspecified,
        prefix: str | None = None,
        suffix: str | None = None,
        x_label: str | None = None,
        y_label: str | None = None,
        raises: bool = True,
        recursive: bool = True,
        strict: bool = False,
        ignore_eq: bool = True,
        comparers: Registry | None = None,
        **options: Any
) -> str | None:
    """
    This is identical to :func:`~testfixtures.compare`, but with ``ignore=True``
    automatically set to make comparing django :class:`~django.db.models.Model`
    instances easier.
    """
    return base_compare(
        *args,
        x=x,
        y=y,
        expected=expected,
        actual=actual,
        prefix=prefix,
        suffix=suffix,
        x_label=x_label,
        y_label=y_label,
        raises=raises,
        recursive=recursive,
        strict=strict,
        ignore_eq=ignore_eq,
        comparers=comparers,
        **options
    )



from operator import setitem
from typing import Any, Callable, Tuple, TypeAlias, Literal

from testfixtures import not_there


Setter: TypeAlias = Callable[[object, str, Any], None] | Callable[[Any, int, Any], None] | None
Key: TypeAlias = Tuple[int, Setter, str | int | None]

class Resolved:

    def __init__(self, container: Any, setter: Setter, name: Any, found: Any):
        self.container: Any = container
        self.setter = setter
        self.name = name
        self.found: Any = found

    def key(self) -> Key:
        return id(self.container), self.setter, self.name

    def __repr__(self) -> str:
        return f'<Resolved: {self.found}>'


def resolve(dotted_name: str, container: Any | None = None, sep: str = '.') -> Resolved:
    names = dotted_name.split(sep)
    used = names.pop(0)
    found: Any
    if container is None:
        found = __import__(used)
        container = found
    else:
        assert not used, 'Absolute traversal not allowed when container supplied'
        used = ''
        found = container
    setter: Setter = None
    name: Any = None
    for name in names:
        container = found
        used += '.' + name
        try:
            found = getattr(found, name)
            setter = setattr
        except AttributeError:
            try:
                if sep != '.':
                    raise ImportError
                __import__(used)
            except ImportError:
                setter = setitem
                try:
                    found = found[name]
                except KeyError:
                    found = not_there
                except TypeError:
                    try:
                        name = int(name)
                    except ValueError:
                        setter = setattr
                        found = not_there
                    else:
                        found = found[name]
            else:
                found = getattr(found, name)
                setter = getattr
            if found is not_there:
                break
    return Resolved(container, setter, name, found)


class _Reference:

    @classmethod
    def classmethod(cls) -> None:  # pragma: no cover
        pass

    @staticmethod
    def staticmethod() -> None:  # pragma: no cover
        pass


class_type = type(_Reference)
classmethod_type = type(_Reference.classmethod)



from django.core.exceptions import ValidationError

from testfixtures import ShouldRaise
from testfixtures.shouldraise import ShouldAssert


class TestShouldRaiseWithValidatorErrors:

    def test_as_expected(self):
        with ShouldRaise(ValidationError("d'oh")):
            raise ValidationError("d'oh")

    def test_not_as_expected(self):
        message = (
            'ValidationError(["d\'oh"]) (expected) != '
            'ValidationError([\'nuts\']) (raised)'
        )
        with ShouldAssert(message):
            with ShouldRaise(ValidationError("d'oh")):
                raise ValidationError("nuts")



"""
A facade for either :mod:`unittest.mock` or its `rolling backport`__, if it is
installed, with a preference for the latter as it may well have newer functionality
and bugfixes.

The facade also contains any bugfixes that are critical to the operation of
functionality provided by testfixtures.

__ https://mock.readthedocs.io
"""
import sys

backport_version: tuple[int, int, int] | None

try:
    from mock import *
    from mock.mock import _Call, _Sentinel
    from mock.mock import call as mock_call
    from mock import version_info as backport_version
except ImportError:
    backport_version = None
    class MockCall:
        pass
    mock_call = MockCall()  # type: ignore[assignment]
    from unittest.mock import *  # type: ignore[assignment]
    from unittest.mock import _Call, _Sentinel  # type: ignore[assignment]


has_backport = backport_version is not None

if not (
        (has_backport and backport_version[:3] > (2, 0, 0)) or  # type: ignore[index]
        (sys.version_info < (3, 0, 0) and not has_backport) or
        (3, 6, 7) < sys.version_info[:3] < (3, 7, 0) or
        sys.version_info[:3] > (3, 7, 1)
):  # pragma: no cover
    raise ImportError('Please upgrade Python (you have {}) or Mock Backport (You have {})'.format(
        sys.version_info, backport_version
    ))
parent_name = '_mock_parent'



from operator import getitem

from testfixtures import (
    Replacer,
    Replace,
    ShouldRaise,
    TempDirectory,
    replace,
    compare,
    not_there,
    replace_in_environ,
    replace_on_class,
    replace_in_module,
    )
from unittest import TestCase

import os

from testfixtures.mock import Mock
from testfixtures.tests import sample1, sample3
from testfixtures.tests import sample2
from .sample1 import z, X
from .sample3 import SOME_CONSTANT
from ..compat import PY_313_PLUS

from warnings import catch_warnings


class TestReplace(TestCase):

    def test_function(self):

        def test_z():
            return 'replacement z'

        compare(sample1.z(), 'original z')

        @replace('testfixtures.tests.sample1.z', test_z)
        def test_something():
            compare(sample1.z(), 'replacement z')

        compare(sample1.z(), 'original z')

        test_something()

        compare(sample1.z(), 'original z')

    def test_class(self):

        OriginalX = sample1.X

        class ReplacementX(sample1.X):
            pass

        self.assertFalse(OriginalX is ReplacementX)
        self.assertTrue(isinstance(sample1.X(), OriginalX))

        @replace('testfixtures.tests.sample1.X', ReplacementX)
        def test_something():
            self.assertFalse(OriginalX is ReplacementX)
            self.assertTrue(isinstance(sample1.X(), ReplacementX))

        self.assertFalse(OriginalX is ReplacementX)
        self.assertTrue(isinstance(sample1.X(), OriginalX))

        test_something()

        self.assertFalse(OriginalX is ReplacementX)
        self.assertTrue(isinstance(sample1.X(), OriginalX))

    def test_method(self):

        def test_y(self):
            return self

        compare(sample1.X().y(), 'original y')

        @replace('testfixtures.tests.sample1.X.y', test_y)
        def test_something():
            self.assertTrue(isinstance(sample1.X().y(), sample1.X))

        compare(sample1.X().y(), 'original y')

        test_something()

        compare(sample1.X().y(), 'original y')

    def test_class_method(self):

        def rMethod(cls):
            return (cls, 1)

        compare(sample1.X().aMethod(), sample1.X)

        @replace('testfixtures.tests.sample1.X.aMethod', rMethod)
        def test_something(r):
            compare(r, rMethod)
            compare(sample1.X().aMethod(), (sample1.X, 1))

        compare(sample1.X().aMethod(), sample1.X)

        test_something()

        compare(sample1.X().aMethod(), sample1.X)

    def test_multiple_replace(self):

        def test_y(self):
            return 'test y'

        def test_z():
            return 'test z'

        compare(sample1.z(), 'original z')
        compare(sample1.X().y(), 'original y')

        @replace('testfixtures.tests.sample1.z', test_z)
        @replace('testfixtures.tests.sample1.X.y', test_y)
        def test_something(passed_test_y, passed_test_z):
            compare(test_z, passed_test_z)
            compare(test_y, passed_test_y)
            compare(sample1.z(), 'test z')
            compare(sample1.X().y(), 'test y')

        compare(sample1.z(), 'original z')
        compare(sample1.X().y(), 'original y')

        test_something()

        compare(sample1.z(), 'original z')
        compare(sample1.X().y(), 'original y')

    def test_gotcha(self):
        # Just because you replace an object in one context,
        # doesn't meant that it's replaced in all contexts!

        def test_z():
            return 'test z'

        compare(sample1.z(), 'original z')
        compare(sample2.z(), 'original z')

        @replace('testfixtures.tests.sample1.z', test_z)
        def test_something():
            compare(sample1.z(), 'test z')
            compare(sample2.z(), 'original z')

        compare(sample1.z(), 'original z')
        compare(sample2.z(), 'original z')

        test_something()

        compare(sample1.z(), 'original z')
        compare(sample2.z(), 'original z')

    def test_raises(self):

        def test_z():
            return 'replacement z'

        compare(sample1.z(), 'original z')

        @replace('testfixtures.tests.sample1.z', test_z)
        def test_something():
            compare(sample1.z(), 'replacement z')
            raise Exception()

        compare(sample1.z(), 'original z')

        with ShouldRaise():
            test_something()

        compare(sample1.z(), 'original z')

    def test_want_replacement(self):

        o = object()

        @replace('testfixtures.tests.sample1.z', o)
        def test_something(r):
            self.assertTrue(r is o)
            self.assertTrue(sample1.z is o)

        test_something()

    def test_not_there(self):

        o = object()

        @replace('testfixtures.tests.sample1.bad', o)
        def test_something(r):
            pass  # pragma: no cover

        with ShouldRaise(AttributeError("Original 'bad' not found")):
            test_something()

    def test_not_there_ok(self):

        o = object()

        @replace('testfixtures.tests.sample1.bad', o, strict=False)
        def test_something(r):
            self.assertTrue(r is o)
            self.assertTrue(sample1.bad is o)

        test_something()

    def test_replace_dict(self):

        from testfixtures.tests.sample1 import some_dict

        original = some_dict['key']
        replacement = object()

        @replace('testfixtures.tests.sample1.some_dict.key', replacement)
        def test_something(obj):
            self.assertTrue(obj is replacement)
            self.assertTrue(some_dict['key'] is replacement)

        test_something()

        self.assertTrue(some_dict['key'] is original)

    def test_replace_delattr(self):

        from testfixtures.tests import sample1

        @replace('testfixtures.tests.sample1.some_dict', not_there)
        def test_something(obj):
            self.assertFalse(hasattr(sample1, 'some_dict'))

        test_something()

        self.assertEqual(sample1.some_dict,
                         {'complex_key': [1, 2, 3], 'key': 'value'})

    def test_replace_delattr_not_there(self):

        @replace('testfixtures.tests.sample1.foo', not_there)
        def test_something(obj):
            pass  # pragma: no cover

        with ShouldRaise(AttributeError("Original 'foo' not found")):
            test_something()

    def test_replace_delattr_not_there_not_strict(self):

        from testfixtures.tests import sample1

        @replace('testfixtures.tests.sample1.foo',
                 not_there, strict=False)
        def test_something(obj):
            self.assertFalse(hasattr(sample1, 'foo'))

        test_something()

    def test_replace_delattr_not_there_restored(self):

        from testfixtures.tests import sample1

        @replace('testfixtures.tests.sample1.foo',
                 not_there, strict=False)
        def test_something(obj):
            sample1.foo = 'bar'

        test_something()
        self.assertFalse(hasattr(sample1, 'foo'))

    def test_replace_delattr_cant_remove(self):
        message = "<class 'datetime.datetime'> has __dict__ but 'today' is not in it"
        with Replacer() as r:
            with ShouldRaise(AttributeError(message)):
                r.replace('datetime.datetime.today', not_there)

    def test_replace_delattr_cant_remove_not_strict(self):
        message = "cannot set 'today' attribute of immutable type 'datetime.datetime'"
        with Replacer() as r:
            with ShouldRaise(TypeError(message)):
                r.replace('datetime.datetime.today', not_there, strict=False)

    def test_replace_dict_remove_key(self):

        from testfixtures.tests.sample1 import some_dict

        @replace('testfixtures.tests.sample1.some_dict.key', not_there)
        def test_something(obj):
            self.assertFalse('key' in some_dict)

        test_something()

        self.assertEqual(sorted(some_dict.keys()), ['complex_key', 'key'])

    def test_replace_dict_remove_key_not_there(self):

        from testfixtures.tests.sample1 import some_dict

        @replace('testfixtures.tests.sample1.some_dict.badkey', not_there)
        def test_something(obj):
            self.assertFalse('badkey' in some_dict)  # pragma: no cover

        with ShouldRaise(AttributeError("Original 'badkey' not found")):
            test_something()

        self.assertEqual(sorted(some_dict.keys()), ['complex_key', 'key'])

    def test_replace_dict_remove_key_not_there_not_strict(self):

        from testfixtures.tests.sample1 import some_dict

        @replace('testfixtures.tests.sample1.some_dict.badkey',
                 not_there, strict=False)
        def test_something(obj):
            self.assertFalse('badkey' in some_dict)

        test_something()

        self.assertEqual(sorted(some_dict.keys()), ['complex_key', 'key'])

    def test_replace_dict_ensure_key_not_there_restored(self):

        from testfixtures.tests.sample1 import some_dict

        @replace('testfixtures.tests.sample1.some_dict.badkey',
                 not_there, strict=False)
        def test_something(obj):
            some_dict['badkey'] = 'some test value'

        test_something()

        self.assertEqual(sorted(some_dict.keys()), ['complex_key', 'key'])

    def test_replace_dict_not_there(self):

        from testfixtures.tests.sample1 import some_dict

        replacement = object()

        @replace('testfixtures.tests.sample1.some_dict.key2',
                 replacement,
                 strict=False)
        def test_something(obj):
            self.assertTrue(obj is replacement)
            self.assertTrue(some_dict['key2'] is replacement)

        test_something()

        self.assertEqual(sorted(some_dict.keys()), ['complex_key', 'key'])

    def test_replace_dict_not_there_empty_string(self):

        from testfixtures.tests.sample1 import some_dict

        @replace('testfixtures.tests.sample1.some_dict.key2', '', strict=False)
        def test_something():
            self.assertEqual(some_dict['key2'], '')

        test_something()

        self.assertEqual(sorted(some_dict.keys()), ['complex_key', 'key'])

    def test_replace_complex(self):

        from testfixtures.tests.sample1 import some_dict

        original = some_dict['complex_key'][1]
        replacement = object()

        @replace('testfixtures.tests.sample1.some_dict.complex_key.1',
                 replacement)
        def test_something(obj):
            self.assertTrue(obj is replacement)
            self.assertEqual(some_dict['complex_key'], [1, obj, 3])

        test_something()

        self.assertEqual(some_dict['complex_key'], [1, 2, 3])

        self.assertTrue(original is some_dict['complex_key'][1])

    def test_replacer_del(self):
        r = Replacer()
        r.replace('testfixtures.tests.sample1.left_behind',
                  object(), strict=False)
        with catch_warnings(record=True) as w:
            del r
            self.assertTrue(len(w), 1)
            compare(str(w[0].message),
                    "Replacer deleted without being restored, originals left:"
                    " {'testfixtures.tests.sample1.left_behind': <Resolved: <not_there>>}")

    def test_multiple_replaces(self):
        orig = os.path.sep
        with Replacer() as r:
            r.replace('os.path.sep', '$')
            compare(os.path.sep, '$')
            r.replace('os.path.sep', '=')
            compare(os.path.sep, '=')
        compare(orig, os.path.sep)

    def test_sub_module_import(self):
        with TempDirectory() as dir:
            dir.write('module/__init__.py', b'')
            dir.write('module/submodule.py', b'def foo(): return "foo"')

            with Replacer() as r:
                r.replace('sys.path', [dir.path])

                def bar():
                    return "bar"
                # now test

                r.replace('module.submodule.foo', bar)

                from module.submodule import foo  # type: ignore[import-not-found]
                compare(foo(), "bar")

    def test_staticmethod(self):
        compare(sample1.X.bMethod(), 2)
        with Replacer() as r:
            r.replace('testfixtures.tests.sample1.X.bMethod', lambda: 1)
            compare(sample1.X.bMethod(), 1)
        compare(sample1.X.bMethod(), 2)

    def test_use_as_cleanup(self):
        def test_z():
            return 'replacement z'

        compare(sample1.z(), 'original z')
        replace = Replacer()
        compare(sample1.z(), 'original z')
        replace('testfixtures.tests.sample1.z', test_z)
        cleanup = replace.restore
        try:
            compare(sample1.z(), 'replacement z')
        finally:
            cleanup()
        compare(sample1.z(), 'original z')

    def test_replace_context_manager(self):
        def test_z():
            return 'replacement z'

        compare(sample1.z(), 'original z')

        with Replace('testfixtures.tests.sample1.z', test_z) as z:
            compare(z(), 'replacement z')
            compare(sample1.z(), 'replacement z')

        compare(sample1.z(), 'original z')

    def test_multiple_context_managers(self):

        def test_y(self):
            return 'test y'

        def test_z():
            return 'test z'

        compare(sample1.z(), 'original z')
        compare(sample1.X().y(), 'original y')

        with Replacer() as replace:
            z = replace('testfixtures.tests.sample1.z', test_z)
            y = replace('testfixtures.tests.sample1.X.y', test_y)
            compare(z(), 'test z')
            compare(y, sample1.X.y)
            compare(sample1.X().y(), 'test y')
            compare(sample1.z(), 'test z')
            compare(sample1.X().y(), 'test y')

        compare(sample1.z(), 'original z')
        compare(sample1.X().y(), 'original y')

    def test_context_manager_not_strict(self):
        def test_z():
            return 'replacement z'

        with Replace('testfixtures.tests.sample1.foo', test_z, strict=False):
            compare(sample1.foo(), 'replacement z')

    def test_context_manager_full_spec(self):
        my_dict = {}

        with Replace(my_dict, name='foo', accessor=getitem, replacement=42, strict=False):
            compare(my_dict, expected={'foo': 42})

        compare(my_dict, expected={})

    def test_decorator_full_spec(self):
        my_dict = {}

        @replace(my_dict, name='foo', accessor=getitem, replacement=42, strict=False)
        def test_something():
            compare(my_dict, expected={'foo': 42})

        test_something()

        compare(my_dict, expected={})

    def test_replace_method_full_spec(self):
        my_dict = {}

        with Replacer() as r:
            r.replace(my_dict, name='foo', accessor=getitem, replacement=42, strict=False)
            compare(my_dict, expected={'foo': 42})

        compare(my_dict, expected={})

    def test_context_manager_specified_method(self):
        class SampleClass:

            def method(self, x):
                return x*2

        sample_obj = SampleClass()

        with Replace(SampleClass.method, lambda self, x: x*3,
                     container=SampleClass, name='method', accessor=getattr):
            compare(sample_obj.method(1), expected=3)

        compare(sample_obj.method(1), expected=2)

    def test_decorator_specified_method(self):
        class SampleClass:

            def method(self, x):
                return x*2

        sample_obj = SampleClass()

        @replace(SampleClass.method, lambda self, x: x*3,
                 container=SampleClass, name='method', accessor=getattr)
        def test_something():
            compare(sample_obj.method(1), expected=3)

        test_something()

        compare(sample_obj.method(1), expected=2)

    def test_fully_specified_method(self):
        class SampleClass:

            def method(self, x):
                return x*2

        sample_obj = SampleClass()

        with Replacer() as r:
            r.replace(SampleClass.method, lambda self, x: x*3,
                      container=SampleClass, name='method', accessor=getattr)
            compare(sample_obj.method(1), expected=3)

        compare(sample_obj.method(1), expected=2)

    def test_fully_specified_method_incorrect_name(self):
        class SampleClass:

            def a(self):
                return 1

            def b(self):
                return 2

        sample_obj = SampleClass()
        s_repr = repr(SampleClass)
        a_repr = repr(SampleClass.a)
        b_repr = repr(SampleClass.b)
        replacer = Replacer()
        with ShouldRaise(AssertionError(
                f"<built-in function getattr> of 'b' from {s_repr} gave {b_repr}, expected {a_repr}"
        )):
            replacer(SampleClass.a, lambda self: 3,
                     container=SampleClass, name='b', accessor=getattr)

        compare(sample_obj.a(), expected=1)
        compare(sample_obj.b(), expected=2)

    def test_traverse_from_container(self):
        x = sample1.X()
        with Replace('.X.aMethod', lambda cls: 'FOO', container=sample1):
            compare(x.aMethod(), expected='FOO')
        assert x.aMethod() is sample1.X

    def test_only_relative_traverse_from_container(self):
        x = sample1.X()
        replacer = Replacer()
        with ShouldRaise(AssertionError('Absolute traversal not allowed when container supplied')):
            replacer('foo', object(), container=sample1)

    def test_no_name_and_target_string(self):
        x = sample1.X()
        replacer = Replacer()
        with ShouldRaise(AttributeError("Original 'aMethod' not found")):
            replacer('.X', lambda cls: 'FOO', name='aMethod', container=sample1)
        assert x.aMethod() is sample1.X

    def test_no_accessor_allowed_when_name_not_specified(self):
        my_dict = {}
        replacer = Replacer()
        with ShouldRaise(TypeError('accessor is not used unless name is specified')):
            replacer(container=my_dict, target='.my.key', accessor=getitem, replacement='bar')
        compare(my_dict, expected={})

    def test_dict_and_name(self):
        my_dict = {}

        with Replace(my_dict, name='foo', replacement=42, strict=False):
            compare(my_dict, expected={'foo': 42})

        compare(my_dict, expected={})

    def test_name_and_dict(self):
        environ = {}

        with Replace('.MY_ENV_VAR', 'True', container=environ, strict=False):
            compare(environ, expected={'MY_ENV_VAR': 'True'})

        compare(environ, expected={})

    def test_non_string_target_and_no_name(self):
        my_dict = {}

        replacer = Replacer()
        with ShouldRaise(TypeError('name must be specified when target is not a string')):
            replacer(my_dict, replacement=42, strict=False)

        compare(my_dict, expected={})

    def test_method_on_instance(self):

        class SampleClass:

            def method(self, x):
                return x*2

        sample = SampleClass()

        with Replacer() as replace:
            replace(sample.method, lambda x: x * 3, container=sample, strict=False)
            compare(sample.method(1), expected=3)

        compare(sample.method(1), expected=2)

    def test_class_attribute(self):

        mock = Mock()
        mock.FOO = set()

        class SampleClass:

            FOO = mock.FOO

        sample = SampleClass()

        with Replace(SampleClass.FOO, {'X'}, container=SampleClass, name='FOO'):
            compare(sample.FOO, expected={'X'})
            compare(mock.FOO, expected=set())

        compare(sample.FOO, expected=set())

    def test_function_and_module(self):
        with Replace(z, lambda: 'all new z', container=sample1):
            from .sample1 import z as sample1_z
            from .sample3 import z as sample3_z
            compare(sample1_z(), expected='all new z')
            compare(sample3_z(), expected='original z')

    def test_constant(self):
        with Replace(SOME_CONSTANT, 43, container=sample3, name='SOME_CONSTANT'):
            from .sample3 import SOME_CONSTANT as sample3_some_constant
            compare(sample3_some_constant, expected=43)

    def test_relative_nested_but_not_present(self):
        nested = {'b': [1, 2, 3]}
        with Replacer() as r:
            with ShouldRaise(AttributeError("Original 'a' not found")):
                r(container=nested, target='.a.1', replacement=42)

    def test_alternative_separator(self):
        nested = {'.b': 1, 'c.d': 2, 'e': {'f': 42}}

        with Replacer() as r:
            r(container=nested, target=':.b', replacement=3, sep=':')
            r(container=nested, target=':c.d', replacement=4, sep=':')
            compare(nested, expected={'.b': 3, 'c.d': 4, 'e': {'f': 42}})

        with Replace(container=nested, target=':e:f', replacement=43, sep=':'):
            compare(nested, expected={'.b': 1, 'c.d': 2, 'e': {'f': 43}})

        @replace(container=nested, target=':e:f', replacement=43, sep=':')
        def function():
            compare(nested, expected={'.b': 1, 'c.d': 2, 'e': {'f': 43}})

        function()

    def test_alternative_separator_traversal_not_valid(self):
        nested = {'.b': 1}

        with Replacer() as r:
            with ShouldRaise(AttributeError("Original '.a.1' not found")):
                r(container=nested, target=':.a.1', replacement=3, sep=':')


class TestEnviron:

    def test_key_present(self):
        os.environ['TESTFIXTURES_SAMPLE_KEY_PRESENT'] = 'ORIGINAL'
        with Replacer() as replace:
            replace.in_environ('TESTFIXTURES_SAMPLE_KEY_PRESENT', 'NEW')
            compare(os.environ['TESTFIXTURES_SAMPLE_KEY_PRESENT'], expected='NEW')
        compare(os.environ['TESTFIXTURES_SAMPLE_KEY_PRESENT'], expected='ORIGINAL')

    def test_key_not_present(self):
        assert 'TESTFIXTURES_SAMPLE_KEY_MISSING' not in os.environ
        with Replacer() as replace:
            replace.in_environ('TESTFIXTURES_SAMPLE_KEY_MISSING', 'NEW')
            compare(os.environ['TESTFIXTURES_SAMPLE_KEY_MISSING'], expected='NEW')
        assert 'TESTFIXTURES_SAMPLE_KEY_MISSING' not in os.environ

    def test_non_string_replacement(self):
        with Replacer() as replace:
            replace.in_environ('PORT', 1)
            compare(os.environ['PORT'], expected='1')

    def test_ensure_not_present(self):
        os.environ['TESTFIXTURES_SAMPLE_KEY_PRESENT'] = 'ORIGINAL'
        with Replacer() as replace:
            replace.in_environ('TESTFIXTURES_SAMPLE_KEY_PRESENT', not_there)
            assert 'TESTFIXTURES_SAMPLE_KEY_PRESENT' not in os.environ


class TestOnClass:

    def test_method_on_class(self):

        class SampleClass:

            def method(self, x):
                return x*2

        original = SampleClass.__dict__['method']
        sample = SampleClass()

        with Replacer() as replace:
            replace.on_class(SampleClass.method, lambda self, x: x*3)
            compare(sample.method(1), expected=3)

        compare(sample.method(1), expected=2)
        assert SampleClass.__dict__['method'] is original

    def test_method_on_subclass(self):

        class SampleClass:

            def method_a(self, x):
                return x*2

        class SampleSubClass(SampleClass):
            """
            Some doc!
            """

            def method_b(self, x):
                return x*3

        original_a = SampleClass.__dict__['method_a']
        original_b = SampleSubClass.__dict__['method_b']
        sample = SampleSubClass()

        with Replacer() as replace:
            replace.on_class(SampleSubClass.method_a, lambda self, x: x*4)
            replace.on_class(SampleSubClass.method_b, lambda self, x: x*5)
            compare(sample.method_a(1), expected=4)
            compare(sample.method_b(1), expected=5)

        compare(sample.method_a(1), expected=2)
        compare(sample.method_b(1), expected=3)
        assert SampleClass.__dict__['method_a'] is original_a
        assert SampleSubClass.__dict__['method_b'] is original_b
        assert 'method_a' not in SampleSubClass.__dict__

    def test_attributes_on_class(self):

        class SampleClass:
            x = 1
            y = 'a'

        sample = SampleClass()

        with Replacer() as replace:
            # without names, we get a useful errors:
            with ShouldRaise(TypeError("attribute named 'x' must be a method")):
                replace.on_class(SampleClass.x, 2, name='x')
            with ShouldRaise(TypeError("attribute must be a method")):
                replace.on_class(SampleClass.y, 'b')

            # okay, so we'll use the full form:
            replace(SampleClass.x, 2, container=SampleClass, name='x')
            replace(SampleClass.y, 'b', container=SampleClass, name='y')
            compare(sample.x, expected=2)
            compare(sample.y, expected='b')

        compare(sample.x, expected=1)
        compare(sample.y, expected='a')

    def test_method_on_instance(self):

        class SampleClass:

            def method(self, x):
                return x*2

        original = SampleClass.__dict__['method']
        sample = SampleClass()

        with Replacer() as replace:
            with ShouldRaise(AttributeError):
                replace.on_class(sample.method, lambda self, x: x*3)

            # ...so use explicit and non-strict:
            replace(sample.method, lambda x: x * 3, container=sample, strict=False)

            compare(sample.method(1), expected=3)

        compare(sample.method(1), expected=2)
        assert SampleClass.__dict__['method'] is original

    def test_badly_decorated_method(self):

        def bad(f):
            def inner(self, x):
                return f(self, x)
            return inner

        class SampleClass:

            @bad
            def method(self, x):
                return x*2

        original = SampleClass.__dict__['method']
        sample = SampleClass()

        with Replacer() as replace:

            # without the name, we get a useful error:
            with ShouldRaise(AttributeError(
                    f"could not find container of {SampleClass.method} using name 'inner'"
            )):
                replace.on_class(SampleClass.method, lambda self_, x: x*3)

            assert SampleClass.__dict__['method'] is original
            replace.on_class(SampleClass.method, lambda self_, x: x*3, name='method')
            compare(sample.method(1), expected=3)

        compare(sample.method(1), expected=2)
        assert SampleClass.__dict__['method'] is original

    def test_classmethod(self):

        class SampleClass:

            @classmethod
            def method(cls, x):
                return x*2

        original = SampleClass.__dict__['method']

        with Replacer() as replace:
            replace.on_class(SampleClass.method, classmethod(lambda cls, x: x*3))
            compare(SampleClass.method(1), expected=3)

        compare(SampleClass.method(1), expected=2)
        assert SampleClass.__dict__['method'] is original

    def test_staticmethod(self):

        class SampleClass:

            @staticmethod
            def method(x):
                return x*2

        original = SampleClass.__dict__['method']

        with Replacer() as replace:
            replace.on_class(SampleClass.method, lambda x: x*3)
            compare(SampleClass.method(1), expected=3)

        compare(SampleClass.method(1), expected=2)
        assert SampleClass.__dict__['method'] is original

    def test_method_on_class_in_module(self):
        sample = X()
        original = X.__dict__['y']

        with Replacer() as replace:
            replace.on_class(X.y, lambda self_: 'replacement y')
            compare(sample.y(), expected='replacement y')

        compare(sample.y(), expected='original y')
        assert X.__dict__['y'] is original

    def test_method_on_instance_in_module(self):

        sample = X()
        original = X.__dict__['y']

        with Replacer() as replace:
            replace(sample.y, lambda: 'replacement y', container=sample, strict=False)
            compare(sample.y(), expected='replacement y')

        compare(sample.y(), expected='original y')
        assert X.__dict__['y'] is original

    def test_classmethod_on_class_in_module(self):

        original = X.__dict__['aMethod']

        with Replacer() as replace:
            replace.on_class(X.aMethod, classmethod(lambda cls: (cls, cls)))
            compare(X.aMethod(), expected=(X, X))

        compare(X.aMethod(), expected=X)
        assert X.__dict__['aMethod'] is original

    def test_classmethod_on_instance_in_module(self):

        sample = X()
        original = X.__dict__['aMethod']

        with Replacer() as replace:
            replace.on_class(sample.aMethod, classmethod(lambda cls: (cls, cls)))
            compare(sample.aMethod(), expected=(X, X))

        compare(sample.aMethod(), expected=X)
        assert X.__dict__['aMethod'] is original

    def test_staticmethod_on_class_in_module(self):

        original = X.__dict__['bMethod']

        with Replacer() as replace:
            replace.on_class(X.bMethod, lambda: 3)
            compare(X.bMethod(), expected=3)

        compare(X.bMethod(), expected=2)
        assert X.__dict__['bMethod'] is original

    def test_staticmethod_on_instance_in_module(self):

        original = X.__dict__['bMethod']
        sample = X()

        with Replacer() as replace:
            replace(sample.bMethod, lambda: 3, container=sample, strict=False)
            compare(sample.bMethod(), expected=3)

        compare(X.bMethod(), expected=2)
        assert X.__dict__['bMethod'] is original

    def test_multiple_methods_on_class(self):
        original_y = X.y
        original_a_result = X.aMethod()
        with Replacer() as replace:
            # y = replace('testfixtures.tests.sample1.X.y', Mock())
            # y.return_value = 'mock y'
            # aMethod = replace('testfixtures.tests.sample1.X.aMethod', Mock())
            # aMethod.return_value = 'mock method'
            replace.on_class(X.y, lambda self: 'mock y')
            replace.on_class(X.aMethod, lambda cls: 'mock method')
            x = X()
            compare(x.y(), expected='mock y')
            compare(x.aMethod(), expected='mock method')

        assert X.y is original_y
        assert X.aMethod() is original_a_result


class TestInModule:

    def test_function_guess_module(self):
        with Replacer() as replace:
            replace.in_module(z, lambda: 'all new z')
            from .sample1 import z as sample1_z
            from .sample3 import z as sample3_z
            compare(sample1_z(), expected='all new z')
            compare(sample3_z(), expected='original z')
        from .sample1 import z as sample1_z
        from .sample3 import z as sample3_z
        compare(sample1_z(), expected='original z')
        compare(sample3_z(), expected='original z')

    def test_function_explict_module(self):
        with Replacer() as replace:
            replace.in_module(z, lambda: 'all new z', module=sample3)
            from .sample1 import z as sample1_z
            from .sample3 import z as sample3_z
            compare(sample1_z(), expected='original z')
            compare(sample3_z(), expected='all new z')
        from .sample1 import z as sample1_z
        from .sample3 import z as sample3_z
        compare(sample1_z(), expected='original z')
        compare(sample3_z(), expected='original z')

    def test_constant(self):
        replace = Replacer()
        with ShouldRaise(AttributeError("'int' object has no attribute '__module__'")):
            replace.in_module(SOME_CONSTANT, 43)
        from .sample3 import SOME_CONSTANT as sample3_some_constant
        compare(sample3_some_constant, expected=42)


class TestConvenience:

    def test_environ(self):
        os.environ['TESTFIXTURES_SAMPLE_KEY_PRESENT'] = 'ORIGINAL'
        with replace_in_environ('TESTFIXTURES_SAMPLE_KEY_PRESENT', 'NEW'):
            compare(os.environ['TESTFIXTURES_SAMPLE_KEY_PRESENT'], expected='NEW')
        compare(os.environ['TESTFIXTURES_SAMPLE_KEY_PRESENT'], expected='ORIGINAL')

    def test_on_class(self):

        class SampleClass:

            def method(self, x):
                return x*2

        sample = SampleClass()

        with replace_on_class(SampleClass.method, lambda self, x: x*3, name='method'):
            compare(sample.method(1), expected=3)

        compare(sample.method(1), expected=2)

    def test_in_module(self):
        with replace_in_module(z, lambda: 'all new z', module=sample3):
            from .sample1 import z as sample1_z
            from .sample3 import z as sample3_z
            compare(sample1_z(), expected='original z')
            compare(sample3_z(), expected='all new z')
        from .sample1 import z as sample1_z
        from .sample3 import z as sample3_z
        compare(sample1_z(), expected='original z')
        compare(sample3_z(), expected='original z')


class OriginA:

    bar = 13

    def __init__(self):
        self.foo = 43

    def method(self, x):
        return x * 2

    original = method


class UseB(OriginA):
    pass


class UseC(OriginA):
    pass


class OriginD:

    def __init__(self, attrs):
        self.attrs = attrs

    def __getattr__(self, item):
        return self.attrs[item]


class OriginE:
    __slots__ = 'attr',

    def method(self, x):
        return x * 2

    original = method


class UseF(OriginE):
    __slots__ = 'attr',


class UseG(OriginE):
    __slots__ = 'attr',


def check_originals_not_modified():
    assert OriginA.__dict__['method'] is OriginA.original
    assert OriginA.__dict__['bar'] == 13
    assert 'method' not in UseB.__dict__
    assert 'method' not in UseC.__dict__
    assert OriginE.method is OriginE.original


def check_behaviour_is_unchanged(*callables):
    for c in callables:
        compare(c(1), expected=2, prefix=repr(c))


class TestReplaceWithInterestingOriginsStrict:

    strict = True

    def test_class_attribute_on_class(self):
        sample_a = OriginA()
        sample_b = UseB()
        sample_c = UseC()

        with Replace(OriginA.bar, name='bar', replacement=31, strict=self.strict, container=OriginA):
            compare(sample_a.bar, expected=31)
            compare(sample_b.bar, expected=31)
            compare(sample_c.bar, expected=31)

        compare(sample_a.bar, expected=13)
        compare(sample_b.bar, expected=13)
        compare(sample_c.bar, expected=13)

        check_originals_not_modified()

    def test_class_attribute_on_subclass(self):
        sample_a = OriginA()
        sample_b = UseB()
        sample_c = UseC()

        replace_ = Replacer()
        with ShouldRaise(AttributeError(f"{UseB!r} has __dict__ but 'bar' is not in it")):
            replace_(UseB.bar, name='bar', replacement=31, strict=self.strict, container=UseB)

        compare(sample_a.bar, expected=13)
        compare(sample_b.bar, expected=13)
        compare(sample_c.bar, expected=13)

        check_originals_not_modified()

    def test_method_on_subclass(self):

        sample_a = OriginA()
        sample_b = UseB()
        sample_c = UseC()

        replace_ = Replacer()
        with ShouldRaise(AttributeError(f"{UseB!r} has __dict__ but 'method' is not in it")):
            replace_(
                UseB.method, lambda self_, x: x*4, name='method', container=UseB, strict=self.strict
            )

        check_behaviour_is_unchanged(sample_a.method, sample_b.method, sample_c.method)
        check_originals_not_modified()

    def test_class_attribute_on_instance_of_class(self):
        obj = OriginA()
        bar = obj.bar

        replace_ = Replacer()
        with ShouldRaise(AttributeError(f"{obj!r} has __dict__ but 'bar' is not in it")):
            replace_(obj, name='bar', replacement=31, strict=self.strict)
        compare(obj.bar, expected=13)

        assert obj.bar is bar

    def test_instance_attribute_on_instance_of_class(self):
        obj = OriginA()
        foo = obj.foo

        with Replace(obj, name='foo', replacement=42, strict=self.strict):
            compare(obj.foo, expected=42)

        assert obj.foo is foo

    def test_method_on_instance_of_class(self):

        sample_a = OriginA()
        sample_b = UseB()
        sample_c = UseC()

        replace = Replacer()
        with ShouldRaise(TypeError(
                "Cannot replace methods on instances with strict=True, "
                "replace on class or use strict=False")
        ):
            replace(
                sample_a.method, lambda self_, x: x*4, name='method', container=sample_a,
                strict=self.strict
            )

        check_behaviour_is_unchanged(sample_a.method, sample_b.method, sample_c.method)
        check_originals_not_modified()

    def test_method_on_instance_of_subclass(self):
        obj = UseB()
        replace = Replacer()
        with ShouldRaise(TypeError(
                "Cannot replace methods on instances with strict=True, "
                "replace on class or use strict=False")
        ):
            replace(
                obj.method, lambda self_, x: x*4, name='method', container=obj, strict=self.strict
            )
        check_behaviour_is_unchanged(obj.method)
        check_originals_not_modified()

    def test_valid_attribute_on_instance_of_slotted_class(self):

        obj = OriginE()
        assert not hasattr(obj, '__dict__')
        obj.attr = 41

        with Replace(obj, name='attr', replacement=42, strict=self.strict):
            compare(obj.attr, expected=42)

        assert obj.attr == 41

    def test_invalid_attribute_on_instance_of_slotted_class(self):
        obj = OriginE()
        assert not hasattr(obj, '__dict__')
        replace_ = Replacer()
        with ShouldRaise(AttributeError("Original 'bad' not found")):
            replace_(obj, name='bad', replacement=42, strict=self.strict)

    def test_method_on_instance_of_slotted_subclass(self):

        sample_e = OriginE()
        sample_f = UseF()
        sample_g = UseG()

        obj = UseF()

        assert not hasattr(obj, '__dict__')

        replace_ = Replacer()
        with ShouldRaise(TypeError(
                "Cannot replace methods on instances with strict=True, "
                "replace on class or use strict=False"
        )):
            replace_(
                obj.method, lambda self_, x: x*4, name='method', container=obj, strict=self.strict
            )

        check_behaviour_is_unchanged(sample_e.method, sample_f.method, sample_g.method)
        check_behaviour_is_unchanged()

    def test_interesting_container(self):
        replace_ = Replacer()
        sample = OriginD({'foo': 'bar'})
        with ShouldRaise(AttributeError(f"{sample!r} has __dict__ but 'foo' is not in it")):
            replace_(sample.foo, 'baz', name='foo', container=sample, strict=self.strict)
        compare(sample.foo, expected='bar')
        assert 'foo' not in sample.__dict__

    def test_mock_and_name(self):
        my_obj = Mock()
        foo = my_obj.foo

        # Mock instances are tricky in that they have a __dict__ but
        # their attributes are not in it:
        replace_ = Replacer()
        with ShouldRaise(AttributeError(f"{my_obj!r} has __dict__ but 'foo' is not in it")):
            replace_(my_obj, name='foo', replacement=42, strict=self.strict)

        assert my_obj.foo is foo


class TestReplaceWithInterestingOriginsNotStrict(TestReplaceWithInterestingOriginsStrict):
    # This subclasses TestReplaceWithInterestingOriginsStrict to ensure we check all the same cases
    strict = False

    def test_class_attribute_on_subclass(self):
        sample_a = OriginA()
        sample_b = UseB()
        sample_c = UseC()

        with Replace(UseB.bar, name='bar', replacement=31, strict=self.strict, container=UseB):
            compare(sample_a.bar, expected=13)
            compare(sample_b.bar, expected=31)
            compare(sample_c.bar, expected=13)

        compare(sample_a.bar, expected=13)
        compare(sample_b.bar, expected=13)
        compare(sample_c.bar, expected=13)

        check_originals_not_modified()

    def test_method_on_subclass(self):

        sample_a = OriginA()
        sample_b = UseB()
        sample_c = UseC()

        with Replace(
                UseB.method, lambda self_, x: x*4, name='method', container=UseB, strict=self.strict
        ):
            compare(sample_b.method(1), expected=4)
            check_behaviour_is_unchanged(sample_a.method, sample_c.method)

        check_behaviour_is_unchanged(sample_a.method, sample_b.method, sample_c.method)
        check_originals_not_modified()

    def test_class_attribute_on_instance_of_class(self):
        obj = OriginA()
        bar = obj.bar

        with Replace(obj, name='bar', replacement=31, strict=self.strict):
            compare(obj.bar, expected=31)

        assert obj.bar is bar

    def test_instance_attribute_on_instance_of_class(self):
        obj = OriginA()
        foo = obj.foo

        with Replace(obj, name='foo', replacement=42, strict=self.strict):
            compare(obj.foo, expected=42)

        assert obj.foo is foo

    def test_method_on_instance_of_class(self):

        sample_a = OriginA()
        sample_b = UseB()
        sample_c = UseC()

        with Replace(
                sample_a.method, lambda x: x*4, name='method', container=sample_a,
                strict=self.strict
        ):
            compare(sample_a.method(1), expected=4)
            check_behaviour_is_unchanged(sample_b.method, sample_c.method)

        check_behaviour_is_unchanged(sample_a.method, sample_b.method, sample_c.method)
        check_originals_not_modified()

    def test_method_on_instance_of_subclass(self):

        sample_a = OriginA()
        sample_b = UseB()
        sample_c = UseC()

        with Replace(
                sample_b.method, lambda x: x*4, name='method', container=sample_b,
                strict=self.strict
        ):
            compare(sample_b.method(1), expected=4)
            check_behaviour_is_unchanged(sample_a.method, sample_c.method)

        check_behaviour_is_unchanged(sample_a.method, sample_b.method, sample_c.method)
        check_originals_not_modified()

    def test_invalid_attribute_on_instance_of_slotted_class(self):
        obj = OriginE()
        assert not hasattr(obj, '__dict__')
        replace_ = Replacer()
        message = "'OriginE' object has no attribute 'bad'"
        if PY_313_PLUS:
            message += " and no __dict__ for setting new attributes"
        with ShouldRaise(AttributeError(message)):
            replace_(obj, name='bad', replacement=42, strict=self.strict)

    def test_method_on_instance_of_slotted_subclass(self):

        sample_e = OriginE()
        sample_f = UseF()
        sample_g = UseG()

        obj = UseF()

        assert not hasattr(obj, '__dict__')

        replace_ = Replacer()
        with ShouldRaise(AttributeError("'UseF' object attribute 'method' is read-only")):
            replace_(
                obj.method, lambda self_, x: x*4, name='method', container=obj, strict=self.strict
            )

        check_behaviour_is_unchanged(sample_e.method, sample_f.method, sample_g.method)
        check_behaviour_is_unchanged()

    def test_interesting_container(self):
        sample = OriginD({'foo': 'bar'})
        with Replace(sample.foo, 'baz', name='foo', container=sample, strict=self.strict):
            compare(sample.foo, expected='baz')
        compare(sample.foo, expected='bar')
        assert 'foo' not in sample.__dict__

    def test_mock_and_name(self):
        my_obj = Mock()
        foo = my_obj.foo

        assert hasattr(my_obj, 'foo')
        assert getattr(my_obj, 'foo', None)

        with Replace(my_obj, name='foo', replacement=42, strict=self.strict):
            compare(my_obj.foo, expected=42)

        with ShouldRaise(AttributeError('foo')):
            # Mock instances are tricky in that they have a __dict__ but
            # their attributes are not in it, confusing Replace, which deletes
            # the attribute on restore as a result:
            assert my_obj.foo is foo



# Copyright (c) 2008-2014 Simplistix Ltd, 2015 onwards Chris Withers
# See license.txt for license details.

import os

from setuptools import setup, find_packages

name = 'testfixtures'
base_dir = os.path.dirname(__file__)

optional = [
    'django',
    'sybil>=6',
    'twisted'
]

setup(
    name=name,
    version=open(os.path.join(base_dir, name, 'version.txt')).read().strip(),
    author='Chris Withers',
    author_email='chris@simplistix.co.uk',
    license='MIT',
    description=("A collection of helpers and mock objects "
                 "for unit tests and doc tests."),
    long_description=open(os.path.join(base_dir, 'README.rst')).read(),
    url='https://github.com/Simplistix/testfixtures',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
    ],
    packages=find_packages(),
    zip_safe=False,
    package_data={'testfixtures': ['py.typed', 'version.txt']},
    include_package_data=True,
    python_requires=">=3.11",
    extras_require=dict(
        test=['django-stubs',
              'mypy',
              'mypy-zope',
              'pytest>=7.1',
              'pytest-cov',
              'pytest-django',
              'types-mock',
              ]+optional,
        docs=['sphinx', 'furo']+optional,
        build=['setuptools-git', 'wheel', 'twine']
    )
)
